/*
  GM 3800 Crank Simulator
  --------------------------------
  The1 Pcmhacking.net


#define PIN_18X     5
#define PIN_3X      6
#define REF_HI_PIN  4
#define BUTTON_PIN  A0

// Global RPM variable
volatile uint16_t currentRPM = 1000; // start at 1000
// Track the current tooth for 18x
volatile uint8_t  currentTooth = 0;

// For button logic
bool lastButtonState     = false;
uint32_t lastButtonPress = 0; // to block multiple increments too fast

void setup() {
  // Pin 5: 18x "inverted" short pulse
  pinMode(PIN_18X, OUTPUT);
  digitalWrite(PIN_18X, HIGH);  // idle state HIGH

  // Pin 6: 3x hardware PWM (inverted)
  pinMode(PIN_3X, OUTPUT);
  digitalWrite(PIN_3X, LOW);    // initial state

  // Pin 4: REF_HI copy of pin6
  pinMode(REF_HI_PIN, OUTPUT);
  digitalWrite(REF_HI_PIN, LOW);

  // Button
  pinMode(BUTTON_PIN, INPUT_PULLUP);

  // Setup Timer3 for 18x pulses
  setupTimer3_18xInterrupt();

  // Setup Timer4 for 3x 50% wave (inverted) on pin 6
  setupTimer4_3xPWM();

  // Initialize RPM
  setRPM(currentRPM);

  // Enable Timer3 Compare A interrupt (for 18x)
  TIMSK3 |= (1 << OCIE3A);

  // Enable Timer4 Compare A & Overflow interrupts (to mirror pin6 on pin4)
  TIMSK4 |= (1 << OCIE4A) | (1 << TOIE4);
}

void loop() {
  // 1) Check button -> increment RPM in 500 steps
  bool buttonPressed = (digitalRead(BUTTON_PIN) == LOW);
  if (buttonPressed && !lastButtonState) {
    uint32_t now = millis();
    if ((now - lastButtonPress) > 20) {
      currentRPM += 500;
      if (currentRPM > 8000) {
        currentRPM = 500;
      }
      setRPM(currentRPM);
      lastButtonPress = now;
    }
  }
  lastButtonState = buttonPressed;

  // ... no long delay here, so pulses remain accurate
}

// ----------------- Timer3 ISR: 18x -----------------
ISR(TIMER3_COMPA_vect) {
  // Produce a short ~50µs LOW pulse on pin 5
  digitalWrite(PIN_18X, LOW);
  delayMicroseconds(50);
  digitalWrite(PIN_18X, HIGH);

  // Move to next tooth
  currentTooth++;
  if (currentTooth >= 18) {
    currentTooth = 0;
  }
}

// ----------------- Timer4: 3x PWM on Pin 4 -----------------

// Compare A => pin6 transitions from LOW->HIGH => set pin4 HIGH
ISR(TIMER4_COMPA_vect) {
  digitalWrite(REF_HI_PIN, HIGH);
}

// Overflow => pin6 transitions from HIGH->LOW => set pin4 LOW
ISR(TIMER4_OVF_vect) {
  digitalWrite(REF_HI_PIN, LOW);
}

// ----------------- Setup Timer3 for 18x -----------------
void setupTimer3_18xInterrupt() {
  cli();
  TCCR3A = 0;
  TCCR3B = 0;
  // Mode: CTC, top=OCR3A => WGM32=1
  TCCR3B |= (1 << WGM32);
  sei();
}

// ----------------- Setup Timer4 for 3x Inverted 50% PWM -----------------
void setupTimer4_3xPWM() {
  // Fast PWM mode 14 => WGM4x bits: WGM41=1, WGM42=1, WGM43=1 => TOP=ICR4
  // Inverting output on OC4A => COM4A1=1, COM4A0=1
  TCCR4A = 0;
  TCCR4B = 0;

  TCCR4A |= (1 << WGM41);
  TCCR4B |= (1 << WGM42) | (1 << WGM43);

  // Inverting on OC4A
  TCCR4A |= (1 << COM4A1) | (1 << COM4A0);
}

// ----------------- setRPM() - Adjust both 18x & 3x to new RPM -----------------
void setRPM(uint16_t rpm) {
  // Safety clamp
  if (rpm < 100)   rpm = 100;
  if (rpm > 10000) rpm = 10000;
  currentRPM = rpm;

  // (1) Timer3 => 18x pulses
  // toothFreq = 18*(rpm/60) => toothPeriod_us = 60e6/(18*rpm)=3,333,333/rpm
  // prescaler=8 => each tick=0.5µs => OCR3A=(tooth_us/0.5)-1 => multiply by 2
  float toothUs = 3333333.0 / rpm;
  uint32_t ticks = (uint32_t)(2.0 * toothUs + 0.5);

  if (ticks < 50)    ticks = 50;     // clamp for high rpm
  if (ticks > 65535) ticks = 65535;  // clamp for low rpm

  OCR3A = (uint16_t)(ticks - 1);

  // Prescaler=8 => CS31=1
  TCCR3B &= ~((1 << CS32) | (1 << CS31) | (1 << CS30));
  TCCR3B |= (1 << CS31);

  // (2) Timer4 => 3x wave with freq=rpm/20 => period=20e6/rpm us
  // prescaler=64 => each tick=4µs => ICR4=(period_us/4)-1
  float periodUs = 20000000.0 / rpm;
  float ticksF   = periodUs / 4.0;   // /4 => prescaler=64
  if (ticksF < 100)    ticksF = 100;
  if (ticksF > 65535)  ticksF = 65535;

  uint16_t top = (uint16_t)(ticksF - 1.0);
  ICR4  = top;

  // 50% duty => OCR4A=top/2
  // But in inverting mode, that means LOW half, HIGH half
  OCR4A = top / 2;

  // Start Timer4 with prescaler=64 => CS42:CS40=011 => decimal 3
  TCCR4B &= ~((1 << CS42) | (1 << CS41) | (1 << CS40));
  TCCR4B |= (1 << CS41) | (1 << CS40);
}
